from typing import Any

from jinja2 import Environment, BaseLoader

from unipipeline.utils.camel_case import camel_case


def template(definition: str, **kwargs: Any) -> str:
    assert isinstance(definition, str), f"definition must be str. {type(definition)} given"

    env = Environment(loader=BaseLoader())
    env.filters['camel'] = camel_case
    template_ = env.from_string(definition)
    result = template_.render(**kwargs)

    return result



from unipipeline import UniWorker

from example.messages.input_message import InputMessage


class InputWorker(UniWorker):
    def handle_message(self, message: InputMessage) -> None:
        print("!!! InputWorker.handle_message ", message)
        raise NotImplementedError('some error in InputWorker.handle_message')



from typing import Set, Any, Optional
from uuid import UUID

from unipipeline.modules.uni_broker_definition import UniBrokerDefinition
from unipipeline.modules.uni_definition import UniDefinition
from unipipeline.modules.uni_message_definition import UniMessageDefinition
from unipipeline.modules.uni_module_definition import UniModuleDefinition
from unipipeline.modules.uni_waiting_definition import UniWaitingDefinition


class UniWorkerDefinition(UniDefinition):
    id: UUID
    name: str
    broker: UniBrokerDefinition[Any]
    type: Optional[UniModuleDefinition]
    topic: str
    error_topic: str
    error_payload_topic: str
    input_message: UniMessageDefinition
    output_workers: Set[str]
    ack_after_success: bool
    waitings: Set[UniWaitingDefinition]
    external: Optional[str]

    @property
    def marked_as_external(self) -> bool:
        return self.external is not None



from typing import Callable, Any, Set, NamedTuple, List, TYPE_CHECKING, Generic, TypeVar

from unipipeline.modules.uni_echo import UniEcho
from unipipeline.modules.uni_broker_definition import UniBrokerDefinition
from unipipeline.modules.uni_message_meta import UniMessageMeta

if TYPE_CHECKING:
    from unipipeline.modules.uni_mediator import UniMediator


class UniBrokerMessageManager:
    def reject(self) -> None:
        raise NotImplementedError(f'method reject must be specified for class "{type(self).__name__}"')

    def ack(self) -> None:
        raise NotImplementedError(f'method acknowledge must be specified for class "{type(self).__name__}"')


class UniBrokerConsumer(NamedTuple):
    id: str
    group_id: str
    message_handler: Callable[[UniMessageMeta, UniBrokerMessageManager], None]


TContent = TypeVar('TContent')


class UniBroker(Generic[TContent]):
    def __init__(self, mediator: 'UniMediator', definition: UniBrokerDefinition[TContent]) -> None:
        self._uni_definition = definition
        self._uni_mediator = mediator
        self._uni_echo = self._uni_mediator.echo.mk_child(f'broker[{self._uni_definition.name}]')

    def connect(self) -> None:
        raise NotImplementedError(f'method connect must be implemented for {type(self).__name__}')

    def close(self) -> None:
        raise NotImplementedError(f'method close must be implemented for {type(self).__name__}')

    def add_topic_consumer(self, topic: str, consumer: UniBrokerConsumer) -> None:
        raise NotImplementedError(f'method consume must be implemented for {type(self).__name__}')

    def end_consuming(self) -> None:
        raise NotImplementedError(f'method end_consuming must be implemented for {type(self).__name__}')

    def start_consuming(self) -> None:
        raise NotImplementedError(f'method start_consuming must be implemented for {type(self).__name__}')

    def publish(self, topic: str, meta_list: List[UniMessageMeta]) -> None:
        raise NotImplementedError(f'method consume must be implemented for {type(self).__name__}')

    def get_topic_approximate_messages_count(self, topic: str) -> int:
        raise NotImplementedError(f'method get_topic_size must be implemented for {type(self).__name__}')

    def initialize(self, topics: Set[str]) -> None:
        raise NotImplementedError(f'method initialize_topic must be implemented for {type(self).__name__}')

    @property
    def definition(self) -> UniBrokerDefinition[Any]:
        return self._uni_definition

    @property
    def echo(self) -> UniEcho:
        return self._uni_echo



from time import sleep
from typing import Dict, TypeVar, Any, Set, Union, Optional, Type, List

from pydantic import ValidationError

from unipipeline.modules.uni_broker import UniBroker
from unipipeline.modules.uni_config import UniConfig
from unipipeline.modules.uni_cron_job import UniCronJob
from unipipeline.modules.uni_echo import UniEcho
from unipipeline.modules.uni_message import UniMessage
from unipipeline.modules.uni_message_meta import UniMessageMeta
from unipipeline.modules.uni_worker import UniWorker
from unipipeline.modules.uni_worker_definition import UniWorkerDefinition
from unipipeline.utils.sig import soft_interruption

TWorker = TypeVar('TWorker', bound=UniWorker)


class PayloadError(Exception):
    def __init__(self, e: ValidationError) -> None:
        super(PayloadError, self).__init__(f'invalid props: {e}')
        self.validation_error = e


class UniMediator:
    def __init__(self, config: UniConfig) -> None:
        self._config = config

        self._worker_definition_by_type: Dict[Any, UniWorkerDefinition] = dict()
        self._worker_instance_indexes: Dict[str, UniWorker] = dict()
        self._broker_instance_indexes: Dict[str, UniBroker[Any]] = dict()
        self._worker_init_list: Set[str] = set()
        self._worker_initialized_list: Set[str] = set()
        self._waiting_init_list: Set[str] = set()
        self._waiting_initialized_list: Set[str] = set()

        self._consumers_list: Set[str] = set()
        self._brokers_with_topics_to_init: Dict[str, Set[str]] = dict()
        self._brokers_with_topics_initialized: Dict[str, Set[str]] = dict()

        self._message_types: Dict[str, Type[UniMessage]] = dict()

        self._brokers_with_active_consumption: List[UniBroker[Any]] = list()

    @property
    def echo(self) -> UniEcho:
        return self._config.echo

    def set_echo_level(self, level: int) -> None:
        self.echo.level = level

    def get_broker(self, name: str, singleton: bool = True) -> UniBroker:
        if not singleton:
            broker_def = self.config.brokers[name]
            broker_type = broker_def.type.import_class(UniBroker, self.echo)
            br = broker_type(mediator=self, definition=broker_def)
            return br
        if name not in self._broker_instance_indexes:
            self._broker_instance_indexes[name] = self.get_broker(name, singleton=False)
        return self._broker_instance_indexes[name]

    def add_worker_to_consume_list(self, name: str) -> None:
        wd = self._config.workers[name]
        if wd.marked_as_external:
            raise OverflowError(f'your could not use worker "{name}" as consumer. it marked as external "{wd.external}"')
        self._consumers_list.add(name)
        self.echo.log_info(f'added consumer {name}')

    def get_message_type(self, name: str) -> Type[UniMessage]:
        if name in self._message_types:
            return self._message_types[name]

        self._message_types[name] = self.config.messages[name].type.import_class(UniMessage, self.echo)

        return self._message_types[name]

    def send_to(self, worker_name: str, payload: Union[Dict[str, Any], UniMessage], parent_meta: Optional[UniMessageMeta] = None, alone: bool = False) -> None:
        if worker_name not in self._worker_initialized_list:
            raise OverflowError(f'worker {worker_name} was not initialized')

        wd = self._config.workers[worker_name]

        message_type = self.get_message_type(wd.input_message.name)
        try:
            if isinstance(payload, message_type):
                payload_data = payload.dict()
            elif isinstance(payload, dict):
                payload_data = message_type(**payload).dict()
            else:
                raise TypeError(f'data has invalid type.{type(payload).__name__} was given')
        except ValidationError as e:
            raise PayloadError(e)

        br = self.get_broker(wd.broker.name)

        if alone:
            size = br.get_topic_approximate_messages_count(wd.topic)
            if size != 0:
                self.echo.log_info(f'sending to worker "{wd.name}" was skipped, because topic {wd.topic} has messages: {size}>0')
                return

        if parent_meta is not None:
            meta = parent_meta.create_child(payload_data)
        else:
            meta = UniMessageMeta.create_new(payload_data)

        meta_list = [meta]
        br.publish(wd.topic, meta_list)  # TODO: make it list by default
        self.echo.log_info(f"worker {wd.name} sent message to topic '{wd.topic}':: {meta_list}")

    def start_cron(self) -> None:
        cron_jobs = UniCronJob.mk_jobs_list(self.config.cron_tasks.values(), self)
        self.echo.log_debug(f'cron jobs defined: {", ".join(cj.task.name for cj in cron_jobs)}')
        while True:
            delay, jobs = UniCronJob.search_next_tasks(cron_jobs)
            if delay is None:
                return
            self.echo.log_debug(f"sleep {delay} seconds before running the tasks: {[cj.task.name for cj in jobs]}")
            if delay > 0:
                sleep(delay)
            self.echo.log_info(f"run the tasks: {[cj.task.name for cj in jobs]}")
            for cj in jobs:
                cj.send()
            sleep(1.1)  # delay for correct next iteration

    def start_consuming(self) -> None:
        brokers = set()
        for wn in self._consumers_list:
            w = self.get_worker(wn)
            w.consume()
            self.echo.log_info(f'consumer {wn} initialized')
            brokers.add(w.definition.broker.name)

        with soft_interruption(self._handle_interruption, self._handle_force_interruption, self._interruption_err):
            for bn in brokers:
                b = self.get_broker(bn)
                self._brokers_with_active_consumption.append(b)
                self.echo.log_info(f'broker {bn} consuming start')
                b.start_consuming()

    def _interruption_err(self, err: Exception) -> None:
        self.echo.log_error(str(err))
        raise err

    def _handle_force_interruption(self) -> None:
        self.echo.log_warning('force interruption detected')

    def _handle_interruption(self) -> None:
        self.echo.log_warning('interruption detected')
        for b in self._brokers_with_active_consumption:
            self.echo.log_debug(f'broker "{b.definition.name}" was notified about interruption')
            b.end_consuming()
        self._brokers_with_active_consumption = list()
        self.echo.log_info(f'all brokers was notified about interruption')

    def add_worker_to_init_list(self, name: str, no_related: bool) -> None:
        if name not in self._config.workers:
            self.echo.exit_with_error(f'worker "{name}" is not found in config "{self.config.file}"')
        wd = self._config.workers[name]
        self._worker_init_list.add(name)
        for waiting in wd.waitings:
            if waiting.name not in self._waiting_initialized_list:
                self._waiting_init_list.add(waiting.name)
        self.add_broker_topic_to_init(wd.broker.name, wd.topic)
        self.add_broker_topic_to_init(wd.broker.name, wd.error_topic)
        self.add_broker_topic_to_init(wd.broker.name, wd.error_payload_topic)
        if not no_related:
            for wn in wd.output_workers:
                self._worker_init_list.add(wn)
                owd = self._config.workers[wn]
                self.add_broker_topic_to_init(owd.broker.name, owd.topic)

    def add_broker_topic_to_init(self, name: str, topic: str) -> None:
        if name in self._brokers_with_topics_initialized:
            if topic in self._brokers_with_topics_initialized[name]:
                return

        if name not in self._brokers_with_topics_to_init:
            self._brokers_with_topics_to_init[name] = set()

        self._brokers_with_topics_to_init[name].add(topic)

    def initialize(self, create: bool = True) -> None:
        for wn in self._worker_init_list:
            self.echo.log_info(f'initialize :: worker "{wn}"', )
            self._worker_initialized_list.add(wn)
        self._worker_init_list = set()

        for waiting_name in self._waiting_init_list:
            self._config.waitings[waiting_name].wait(self.echo)
            self.echo.log_info(f'initialize :: waiting "{waiting_name}"')
            self._waiting_initialized_list.add(waiting_name)
        self._waiting_init_list = set()

        if create:
            for bn, topics in self._brokers_with_topics_to_init.items():
                b = self.wait_for_broker_connection(bn)

                b.initialize(topics)
                self.echo.log_info(f'initialize :: broker "{b.definition.name}" topics :: {topics}')

                if bn not in self._brokers_with_topics_initialized:
                    self._brokers_with_topics_initialized[bn] = set()
                for topic in topics:
                    self._brokers_with_topics_initialized[bn].add(topic)
            self._brokers_with_topics_to_init = dict()

    def get_worker(self, worker: Union[Type['UniWorker[UniMessage]'], str], singleton: bool = True) -> UniWorker[UniMessage]:
        wd = self._config.get_worker_definition(worker)
        if wd.marked_as_external:
            raise OverflowError(f'worker "{worker}" is external. you could not get it')
        if not singleton or wd.name not in self._worker_instance_indexes:
            assert wd.type is not None
            worker_type = wd.type.import_class(UniWorker, self.echo)
            self.echo.log_info(f'get_worker :: initialized worker "{wd.name}"')
            w = worker_type(definition=wd, mediator=self)
        else:
            return self._worker_instance_indexes[wd.name]
        self._worker_instance_indexes[wd.name] = w
        return w

    @property
    def config(self) -> UniConfig:
        return self._config

    def wait_for_broker_connection(self, name: str) -> UniBroker:
        br = self.get_broker(name)
        for try_count in range(br.definition.retry_max_count):
            try:
                br.connect()
                self.echo.log_info(f'wait_for_broker_connection :: broker {br.definition.name} connected')
                return br
            except ConnectionError as e:
                self.echo.log_info(f'wait_for_broker_connection :: broker {br.definition.name} retry to connect [{try_count}/{br.definition.retry_max_count}] : {e}')
                sleep(br.definition.retry_delay_s)
                continue
        raise ConnectionError(f'unavailable connection to {br.definition.name}')




from typing import Set, List

from unipipeline import UniBroker, UniMessageMeta, UniBrokerConsumer


class LogBroker(UniBroker):

    def connect(self) -> None:
        pass

    def close(self) -> None:
        pass

    def add_topic_consumer(self, topic: str, consumer: UniBrokerConsumer) -> None:
        pass

    def start_consuming(self) -> None:
        pass

    def publish(self, topic: str, meta_list: List[UniMessageMeta]) -> None:
        pass

    def get_topic_approximate_messages_count(self, topic: str) -> int:
        pass

    def initialize(self, topics: Set[str]) -> None:
        pass



import logging
import os
import sys

from unipipeline import Uni
from unipipeline.args import CMD_INIT, CMD_CHECK, CMD_CRON, CMD_PRODUCE, CMD_CONSUME, parse_args, CMD_SCAFFOLD


def run_check(u: Uni, args) -> None:
    u.check(create=False)


def run_scaffold(u: Uni, args) -> None:
    u.check(create=True)


def run_cron(u: Uni, args) -> None:
    u.init_cron()
    u.initialize()
    u.start_cron()


def run_init(u: Uni, args) -> None:
    for wn in args.init_workers:
        u.init_producer_worker(wn)
    u.initialize(everything=len(args.init_workers) == 0, create=True)


def run_consume(u: Uni, args) -> None:
    for wn in args.consume_workers:
        u.init_consumer_worker(wn)
    u.initialize()
    u.start_consuming()


def run_produce(u: Uni, args) -> None:
    u.init_producer_worker(args.produce_worker)
    u.initialize()
    u.send_to(args.produce_worker, args.produce_data, alone=args.produce_alone)


args_cmd_map = {
    CMD_INIT: run_init,
    CMD_CHECK: run_check,
    CMD_SCAFFOLD: run_scaffold,
    CMD_CRON: run_cron,
    CMD_PRODUCE: run_produce,
    CMD_CONSUME: run_consume,
}


def main():
    sys.path.insert(0, os.getcwdb().decode('utf-8'))
    args = parse_args()
    print(args)
    u = Uni(args.config_file, echo_level=logging.DEBUG if args.verbose else None)
    args_cmd_map[args.cmd](u, args)
    u.echo.success('done')




COLOR_MAGENTA = '\u001b[35m'
COLOR_YELLOW = '\u001b[33m'
COLOR_GREEN = '\u001b[32m'
COLOR_WHITE = '\u001b[37m'
COLOR_BLACK = '\u001b[30m'
COLOR_RESET = '\u001b[0m'
COLOR_GRAY = '\u001b[0;37m'
COLOR_BLUE = '\u001b[34m'
COLOR_CYAN = '\u001b[36m'
COLOR_RED = '\u001b[31m'


def color_it(clr: str, msg: str) -> str:
    return f'{clr}{msg}{COLOR_RESET}'



import json
from typing import Optional, Tuple, Any, Dict, List, NamedTuple, Set

from kafka import KafkaProducer, KafkaConsumer  # type: ignore
from kafka.consumer.fetcher import ConsumerRecord  # type: ignore
from pydantic import BaseModel

from unipipeline.modules.uni_broker import UniBroker, UniBrokerMessageManager, UniBrokerConsumer
from unipipeline.modules.uni_message_meta import UniMessageMeta


class UniKafkaBrokerMessageManager(UniBrokerMessageManager):
    def reject(self) -> None:
        pass

    def ack(self) -> None:
        pass


class UniKafkaBrokerConf(BaseModel):
    api_version: List[int]


class UniKafkaBrokerConsumer(NamedTuple):
    kfk_consumer: KafkaConsumer
    consumer: UniBrokerConsumer


class UniKafkaBroker(UniBroker[bytes]):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.conf = UniKafkaBrokerConf(**self.definition.configure_dynamic(dict(
           api_version=None
        )))

        self._bootstrap_servers = self.get_boostrap_servers()

        self._producer: Optional[KafkaProducer] = None

        self._security_conf: Dict[str, Any] = self.get_security_conf()

        self._consumers: List[UniKafkaBrokerConsumer] = list()

        self._consuming_started = False
        self._interrupted = False
        self._in_processing = False

    def end_consuming(self) -> None:    # TODO
        self._end_consuming()

    def _end_consuming(self) -> None:
        self._interrupted = True
        if not self._in_processing:
            for consumer in self._consumers:
                consumer.kfk_consumer.close()

    def get_topic_approximate_messages_count(self, topic: str) -> int:
        return 0  # TODO

    def initialize(self, topics: Set[str]) -> None:
        pass  # TODO

    def get_boostrap_servers(self) -> List[str]:
        raise NotImplementedError(f'method get_boostrap_server must be implemented for {type(self).__name__}')

    def get_security_conf(self) -> Dict[str, Any]:
        raise NotImplementedError(f'method get_security_conf must be implemented for {type(self).__name__}')

    def connect(self) -> None:
        if self._producer is not None:
            return

        self._producer = KafkaProducer(
            bootstrap_servers=self._bootstrap_servers,
            api_version=self.conf.api_version,
            **self._security_conf,
        )

    def close(self) -> None:
        if self._producer is not None:
            self._producer.close()
            self._producer = None
        for consumer in self._consumers:
            consumer.kfk_consumer.close()

    def _serialize_body(self, meta: UniMessageMeta) -> Tuple[bytes, bytes]:
        meta_dumps = self.definition.codec.dumps(meta.dict())
        return str(meta.id).encode('utf8'), bytes(meta_dumps, encoding='utf8')

    def _parse_body(self, msg: ConsumerRecord) -> UniMessageMeta:
        if msg.value.get('parent', None) is not None:
            return UniMessageMeta(**msg.value)
        return UniMessageMeta.create_new(data=msg.value)

    def add_topic_consumer(self, topic: str, consumer: UniBrokerConsumer) -> None:
        kfk_consumer = KafkaConsumer(
            topic,
            api_version=self.conf.api_version,
            bootstrap_servers=self._bootstrap_servers,
            value_deserializer=lambda m: json.loads(m.decode('utf-8')),
            group_id=consumer.group_id,
        )
        self._consumers.append(UniKafkaBrokerConsumer(kfk_consumer, consumer))

    def start_consuming(self) -> None:
        self.connect()
        if len(self._consumers) == 0:
            self.echo.log_warning('no consumers to start')
            return
        if self._consuming_started:
            raise OverflowError('consuming was started')
        self._consuming_started = True
        self._interrupted = False
        self._in_processing = False
        for cnsmr in self._consumers:  # TODO: make it asynchronously
            msg = next(cnsmr.kfk_consumer)
            self._in_processing = True

            meta = self._parse_body(msg)
            manager = UniKafkaBrokerMessageManager()
            cnsmr.consumer.message_handler(meta, manager)
            self._in_processing = False
            if self._interrupted:
                self._end_consuming()
                break

    def publish(self, topic: str, meta_list: List[UniMessageMeta]) -> None:
        self.connect()
        assert self._producer is not None
        for meta in meta_list:
            key, value = self._serialize_body(meta)
            self._producer.send(topic=topic, value=value, key=key)
        self._producer.flush()



from typing import Generic, Type, Any, TypeVar, Optional, Dict, Union
from uuid import uuid4

from unipipeline.modules.uni_broker import UniBrokerMessageManager, UniBrokerConsumer
from unipipeline.modules.uni_message import UniMessage
from unipipeline.modules.uni_message_meta import UniMessageMeta, UniMessageMetaErrTopic
from unipipeline.modules.uni_worker_definition import UniWorkerDefinition

TMessage = TypeVar('TMessage', bound=UniMessage)


class UniPayloadParsingError(Exception):
    def __init__(self, exception: Exception):
        self.parent_exception = exception


class UniWorker(Generic[TMessage]):
    def __init__(
        self,
        definition: UniWorkerDefinition,
        mediator: Any
    ) -> None:
        from unipipeline.modules.uni_mediator import UniMediator
        self._uni_moved = False
        self._uni_consume_initialized = False
        self._uni_payload_cache: Optional[TMessage] = None
        self._uni_current_meta: Optional[UniMessageMeta] = None
        self._uni_current_manager: Optional[UniBrokerMessageManager] = None
        self._uni_definition = definition
        self._uni_mediator: UniMediator = mediator
        self._uni_worker_instances_for_sending: Dict[Type[UniWorker], UniWorker] = dict()
        self._uni_echo = self._uni_mediator.echo.mk_child(f'worker[{self._uni_definition.name}]')
        self._uni_message_type: Type[TMessage] = self._uni_definition.input_message.type.import_class(UniMessage, self._uni_echo)  # type: ignore

    @property
    def message_type(self) -> Type[TMessage]:
        return self._uni_message_type

    def consume(self) -> None:
        if self._uni_consume_initialized:
            raise OverflowError(f'consumer of worker "{self._uni_definition.name}" has already initialized')
        self._uni_consume_initialized = True
        br = self._uni_mediator.get_broker(self._uni_definition.broker.name)

        self._uni_echo.log_info(f"worker {self._uni_definition.name} start consuming")
        br.add_topic_consumer(
            topic=self._uni_definition.topic,
            consumer=UniBrokerConsumer(
                id=f'{self._uni_definition.name}__{uuid4()}',
                group_id=self._uni_definition.name,
                message_handler=self.process_message
            ),
        )

    @property
    def definition(self) -> UniWorkerDefinition:
        return self._uni_definition

    @property
    def meta(self) -> UniMessageMeta:
        assert self._uni_current_meta is not None
        return self._uni_current_meta

    @property
    def manager(self) -> UniBrokerMessageManager:
        assert self._uni_current_manager is not None
        return self._uni_current_manager

    @property
    def payload(self) -> TMessage:
        if self._uni_payload_cache is None:
            try:
                self._uni_payload_cache = self._uni_message_type(**self.meta.payload)
            except Exception as e:
                raise UniPayloadParsingError(e)
        return self._uni_payload_cache

    def send_to(self, worker: Union[Type['UniWorker[TMessage]'], str], data: Any, alone: bool = False) -> None:
        if self._uni_current_meta is None:
            raise ValueError(f'meta was not defined. incorrect usage of function "send_to"')

        wd = self._uni_mediator.config.workers[worker] if isinstance(worker, str) else self._uni_mediator.config.workers_by_class[worker.__name__]

        if wd.name not in self._uni_definition.output_workers:
            raise ValueError(f'worker {wd.name} is not defined in workers->{self._uni_definition.name}->output_workers')

        self._uni_mediator.send_to(wd.name, data, parent_meta=self._uni_current_meta, alone=alone)

    def process_message(self, meta: UniMessageMeta, manager: UniBrokerMessageManager) -> None:
        self._uni_echo.log_debug(f"worker {self._uni_definition.name} message {meta} received")
        self._uni_moved = False
        self._uni_payload_cache = None
        self._uni_current_meta = meta
        self._uni_current_manager = manager

        unsupported_err_topic = False
        if not meta.has_error:
            try:
                self.handle_message(self.payload)
            except UniPayloadParsingError as e:
                self.move_to_error_topic(UniMessageMetaErrTopic.MESSAGE_PAYLOAD_ERR, e)
            except Exception as e:
                self._uni_echo.log_error(str(e))
                self.move_to_error_topic(UniMessageMetaErrTopic.HANDLE_MESSAGE_ERR, e)
        else:
            try:
                assert meta.error is not None  # for mypy needs
                if meta.error.error_topic is UniMessageMetaErrTopic.HANDLE_MESSAGE_ERR:
                    self.handle_error_message_handling(self.payload)
                elif meta.error.error_topic is UniMessageMetaErrTopic.MESSAGE_PAYLOAD_ERR:
                    self.handle_error_message_payload(self.meta, self.manager)
                elif meta.error.error_topic is UniMessageMetaErrTopic.ERROR_HANDLING_ERR:
                    self.handle_error_handling(self.meta, self.manager)
                else:
                    unsupported_err_topic = True
            except Exception as e:
                self._uni_echo.log_error(str(e))
                self.move_to_error_topic(UniMessageMetaErrTopic.ERROR_HANDLING_ERR, e)

        if unsupported_err_topic:
            assert meta.error is not None  # for mypy needs
            err = NotImplementedError(f'{meta.error.error_topic} is not implemented in process_message')
            self._uni_echo.log_error(str(err))
            self.move_to_error_topic(UniMessageMetaErrTopic.SYSTEM_ERR, err)

        if not self._uni_moved and self._uni_definition.ack_after_success:
            manager.ack()

        self._uni_echo.log_debug(f"worker message {meta} processed")
        self._uni_moved = False
        self._uni_current_meta = None
        self._uni_current_manager = None

    def handle_message(self, message: TMessage) -> None:
        raise NotImplementedError(f'method handle_message not implemented for {type(self).__name__}')

    def move_to_error_topic(self, err_topic: UniMessageMetaErrTopic, err: Exception) -> None:
        self._uni_moved = True
        meta = self.meta.create_error_child(err_topic, err)
        br = self._uni_mediator.get_broker(self._uni_definition.broker.name)
        error_topic = self.definition.error_topic
        if error_topic is UniMessageMetaErrTopic.MESSAGE_PAYLOAD_ERR:
            error_topic = self.definition.error_payload_topic
        br.publish(error_topic, [meta])
        self.manager.ack()

    def handle_error_message_handling(self, message: TMessage) -> None:
        pass

    def handle_error_message_payload(self, meta: UniMessageMeta, manager: UniBrokerMessageManager) -> None:
        pass

    def handle_error_handling(self, meta: UniMessageMeta, manager: UniBrokerMessageManager) -> None:
        pass

    def handle_uni_error(self, meta: UniMessageMeta, manager: UniBrokerMessageManager) -> None:
        pass



from typing import Optional, Any, Tuple, TypeVar, Set, List
from urllib.parse import urlparse

from pika import ConnectionParameters, PlainCredentials, BlockingConnection, BasicProperties, spec  # type: ignore
from pika.adapters.blocking_connection import BlockingChannel  # type: ignore
from pydantic import BaseModel

from unipipeline.modules.uni_broker import UniBroker, UniBrokerMessageManager, UniBrokerConsumer
from unipipeline.modules.uni_message import UniMessage
from unipipeline.modules.uni_message_codec import UniMessageCodec
from unipipeline.modules.uni_message_meta import UniMessageMeta
from unipipeline.utils.connection_pool import ConnectionObj, TConnectionObj, connection_pool

BASIC_PROPERTIES__HEADER__COMPRESSION_KEY = 'compression'


TMessage = TypeVar('TMessage', bound=UniMessage)


class UniAmqpBrokerMessageManager(UniBrokerMessageManager):

    def __init__(self, channel: BlockingChannel, method_frame: spec.Basic.Deliver) -> None:
        self._channel = channel
        self._method_frame = method_frame
        self._acknowledged = False

    def reject(self) -> None:
        self._channel.basic_reject(delivery_tag=self._method_frame.delivery_tag, requeue=True)

    def ack(self) -> None:
        if self._acknowledged:
            return
        self._acknowledged = True
        self._channel.basic_ack(delivery_tag=self._method_frame.delivery_tag)


class UniAmqpBrokerConnectionObj(ConnectionObj[BlockingConnection]):

    def __init__(self, params: ConnectionParameters) -> None:
        self._params = params
        self._connection: Optional[BlockingConnection] = None

    def __hash__(self) -> int:
        return hash(f'{self._params.host}{self._params.port}{self._params.credentials.username}{self._params.credentials.password}')

    def get(self) -> TConnectionObj:
        assert self._connection is not None
        return self._connection

    def is_closed(self) -> bool:
        return self._connection is None or self._connection.is_closed

    def connect(self) -> None:
        assert self._connection is None
        self._connection = BlockingConnection(self._params)

    def close(self):
        if self._connection is not None:
            if not self._connection.is_closed:
                self._connection.close()
            self._connection = None


class UniAmqpBrokerConfig(BaseModel):
    exchange_name: str
    heartbeat: int
    blocked_connection_timeout: int
    prefetch: int
    socket_timeout: int
    stack_timeout: int
    exchange_type: str
    durable: bool
    auto_delete: bool
    passive: bool
    is_persistent: bool


class UniAmqpBroker(UniBroker[bytes]):
    def get_topic_approximate_messages_count(self, topic: str) -> int:
        raise NotImplementedError(f"method get_topic_approximate_messages_count must be implemented for class '{type(self).__name__}'")  # TODO

    @classmethod
    def get_connection_uri(cls) -> str:
        raise NotImplementedError(f"cls method get_connection_uri must be implemented for class '{cls.__name__}'")

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.conf = UniAmqpBrokerConfig(**self.definition.configure_dynamic(dict(
            exchange_name="communication",
            heartbeat=600,
            blocked_connection_timeout=300,
            prefetch=1,
            socket_timeout=300,
            stack_timeout=300,
            exchange_type="direct",
            durable=True,
            auto_delete=False,
            passive=False,
            is_persistent=True,
        )))

        self._consumers_count = 0

        self._connection_key = self.get_connection_uri()
        url_params_pr = urlparse(url=self._connection_key)

        self._connector = connection_pool.new_manager(UniAmqpBrokerConnectionObj(ConnectionParameters(
            heartbeat=self.conf.heartbeat,
            blocked_connection_timeout=self.conf.blocked_connection_timeout,
            socket_timeout=self.conf.socket_timeout,
            stack_timeout=self.conf.stack_timeout,
            retry_delay=self.definition.retry_delay_s,
            host=url_params_pr.hostname,
            port=url_params_pr.port,
            credentials=PlainCredentials(url_params_pr.username, url_params_pr.password, erase_on_connect=False),
        )))

        self._channel: Optional[BlockingChannel] = None
        self._consuming_started = False

        self._in_processing = False
        self._interrupted = False

    def initialize(self, topics: Set[str]) -> None:
        ch = self._get_channel()
        ch.exchange_declare(
            exchange=self.conf.exchange_name,
            exchange_type=self.conf.exchange_type,
            passive=self.conf.passive,
            durable=self.conf.durable,
            auto_delete=self.conf.auto_delete,
        )

        ch.basic_qos(prefetch_count=self.conf.prefetch)

        for topic in topics:
            ch.queue_declare(
                queue=topic,
                durable=self.conf.durable,
                auto_delete=self.conf.auto_delete
            )

            ch.queue_bind(queue=topic, exchange=self.conf.exchange_name, routing_key=topic)

    def end_consuming(self) -> None:
        self._end_consuming()

    def _end_consuming(self):
        self._interrupted = True
        if not self._in_processing:
            self._get_channel().stop_consuming()

    def connect(self) -> None:
        self._connector.connect()

    def close(self) -> None:
        self._connector.close()

    def _get_channel(self) -> BlockingChannel:
        if self._channel is not None:
            return self._channel

        self._channel = self._connector.connect().channel()

        return self._channel

    def add_topic_consumer(self, topic: str, consumer: UniBrokerConsumer) -> None:
        if self._consuming_started:
            raise OverflowError('you cannot add consumer dynamically')

        self._consumers_count += 1

        def consumer_wrapper(channel: BlockingChannel, method_frame: spec.Basic.Deliver, properties: BasicProperties, body: bytes) -> None:
            self._in_processing = True
            manager = UniAmqpBrokerMessageManager(channel, method_frame)
            meta = self.parse_body(body, properties)
            consumer.message_handler(meta, manager)
            self._in_processing = False
            if self._interrupted:
                self._end_consuming()

        self._get_channel().basic_consume(
            queue=topic,
            on_message_callback=consumer_wrapper,
            consumer_tag=consumer.id,
        )
        self.echo.log_info(f'added consumer for topic "{topic}" with consumer_tag "{consumer.id}')

    def start_consuming(self) -> None:
        if self._consumers_count == 0:
            self.echo.log_warning('has no consumers to start consuming')
            return
        if self._consuming_started:
            raise OverflowError('you cannot consume twice!')
        self._consuming_started = True
        self._interrupted = False
        self._in_processing = False

        self.echo.log_info(f'start consuming:: has {self._consumers_count} workers')
        self._get_channel().start_consuming()  # blocking operation

    def serialize_body(self, meta: UniMessageMeta) -> Tuple[bytes, BasicProperties]:
        meta_dumps = self.definition.codec.dumps(meta.dict())
        meta_compressed = self.definition.codec.compress(meta_dumps)

        properties = BasicProperties(
            content_type=self.definition.codec.content_type,
            content_encoding='utf-8',
            delivery_mode=2 if self.conf.is_persistent else 0,
            headers={BASIC_PROPERTIES__HEADER__COMPRESSION_KEY: self.definition.codec.compression}
        )
        return meta_compressed, properties

    def parse_body(self, body: bytes, properties: BasicProperties) -> UniMessageMeta:
        codec: UniMessageCodec[Any] = UniMessageCodec(
            compression=properties.headers.get(BASIC_PROPERTIES__HEADER__COMPRESSION_KEY, None),
            content_type=properties.content_type
        )
        body_uncompressed = codec.decompress(body)
        body_json = codec.loads(body_uncompressed)
        return UniMessageMeta(**body_json)

    def publish(self, topic: str, meta_list: List[UniMessageMeta]) -> None:
        for meta in meta_list:
            body, properties = self.serialize_body(meta)
            self._get_channel().basic_publish(
                exchange=self.conf.exchange_name,
                routing_key=topic,
                body=body,
                properties=properties
            )



from typing import Dict, Any, Set, Iterator, Tuple
from uuid import uuid4


COMMON_PROP = "__default__"


class ParseDefinitionError(Exception):
    pass


def parse_definition(
    conf_name: str,
    definitions: Dict[str, Any],
    defaults: Dict[str, Any],
    required_keys: Set[str],
    sys_names: Set[str] = None
) -> Iterator[Tuple[str, Dict[str, Any], Dict[str, Any]]]:
    if sys_names is None:
        sys_names = set()

    if not isinstance(definitions, dict):
        raise ParseDefinitionError(f'definition of {conf_name} has invalid type. must be dict')

    defaults_keys = set(defaults.keys())

    common = definitions.get(COMMON_PROP, dict())

    for name, raw_definition in definitions.items():
        other_props: Dict[str, Any] = dict()
        definition = dict(defaults)

        name = str(name)
        if name.startswith("_"):
            if name == COMMON_PROP:
                continue
            elif name in sys_names:
                pass
            else:
                raise ValueError(f"key '{name}' is not acceptable")

        if not isinstance(raw_definition, dict):
            raise ParseDefinitionError(f'definition of {conf_name}->{name} has invalid type. must be dict')

        raw_definition.update(common)

        for k, v in raw_definition.items():
            if k in defaults_keys:
                definition[k] = v
                vd = defaults.get(k, None)
                if vd is not None and type(vd) != type(v):
                    raise ParseDefinitionError(f'definition of {conf_name}->{name} has invalid key "{k}" type')
            elif k in required_keys:
                definition[k] = v
            else:
                other_props[k] = v

        for rk in required_keys:
            if rk not in definition:
                raise ParseDefinitionError(f'definition of {conf_name}->{name} has no required prop "{rk}"')

        definition["name"] = name
        definition["id"] = uuid4()

        yield name, definition, other_props



from typing import Generic, TypeVar
from uuid import UUID

from unipipeline.modules.uni_definition import UniDefinition
from unipipeline.modules.uni_message_codec import UniMessageCodec
from unipipeline.modules.uni_module_definition import UniModuleDefinition

TContent = TypeVar('TContent')


class UniBrokerDefinition(UniDefinition, Generic[TContent]):
    id: UUID
    type: UniModuleDefinition

    retry_max_count: int
    retry_delay_s: int

    codec: UniMessageCodec[TContent]



from typing import Any, Dict

from pydantic import BaseModel

from unipipeline.utils.parse_definition import ParseDefinitionError


class UniDefinition(BaseModel):
    name: str

    dynamic_props_: Dict[str, Any]

    def configure_dynamic(self, defaults: Dict[str, Any]) -> Dict[str, Any]:
        res = dict()
        for k, v in defaults.items():
            if k in self.dynamic_props_:
                res[k] = self.dynamic_props_[k]
                if v is not None and not isinstance(res[k], type(v)):
                    raise ParseDefinitionError(f'configure broker {self.name}->{k} has invalid property type. must be {type(v)}')
            else:
                res[k] = v
        return res



from typing import Dict, Any, Union, Optional

from unipipeline.modules.uni_broker import UniBroker
from unipipeline.modules.uni_config import UniConfig, UniConfigError
from unipipeline.modules.uni_echo import UniEcho
from unipipeline.modules.uni_mediator import UniMediator, PayloadError
from unipipeline.modules.uni_message import UniMessage
from unipipeline.modules.uni_wating import UniWaiting
from unipipeline.modules.uni_worker import UniWorker
from unipipeline.utils.parse_definition import ParseDefinitionError


class Uni:
    def __init__(self, config: Union[UniConfig, str], echo_level: Optional[Union[str, int]] = None) -> None:
        if isinstance(config, str):
            config = UniConfig(config, echo_level)
        if not isinstance(config, UniConfig):
            raise ValueError(f'invalid config type. {type(config).__name__} was given')
        self._mediator = UniMediator(config)

    @property
    def echo(self) -> UniEcho:
        return self._mediator.echo

    def set_echo_level(self, level: int):
        self._mediator.set_echo_level(level)

    def check(self, create: bool = False) -> None:
        try:
            for waiting_def in self._mediator.config.waitings.values():
                waiting_def.type.import_class(UniWaiting, self.echo, create, create_template_params=waiting_def)

            for broker_def in self._mediator.config.brokers.values():
                broker_def.type.import_class(UniBroker, self.echo, create, create_template_params=broker_def)

            for message_def in self._mediator.config.messages.values():
                message_def.type.import_class(UniMessage, self.echo, create, create_template_params=message_def)

            for worker_def in self._mediator.config.workers.values():
                if worker_def.marked_as_external:
                    continue
                assert worker_def.type is not None
                worker_def.type.import_class(UniWorker, self.echo, create, create_template_params=worker_def)

        except (ParseDefinitionError, UniConfigError) as e:
            print(f"ERROR: {e}")
            exit(1)

    def start_cron(self) -> None:
        try:
            self._mediator.start_cron()
        except KeyboardInterrupt:
            self.echo.log_warning('interrupted')
            exit(0)

    def initialize(self, everything: bool = False, create: bool = True) -> None:
        if everything:
            for wn in self._mediator.config.workers.keys():
                self._mediator.add_worker_to_init_list(wn, no_related=True)
        self._mediator.initialize(create=create)

    def init_cron(self) -> None:
        for task in self._mediator.config.cron_tasks.values():
            self._mediator.add_worker_to_init_list(task.worker.name, no_related=True)

    def init_producer_worker(self, name: str) -> None:
        self._mediator.add_worker_to_init_list(name, no_related=True)

    def init_consumer_worker(self, name: str) -> None:
        self._mediator.add_worker_to_init_list(name, no_related=False)
        self._mediator.add_worker_to_consume_list(name)

    def send_to(self, name: str, data: Union[Dict[str, Any], UniMessage], alone: bool = False) -> None:
        try:
            self._mediator.send_to(name, data, alone=alone)
        except PayloadError as e:
            self.echo.exit_with_error(f'invalid props in message: {e.validation_error}')

    def start_consuming(self) -> None:
        try:
            self._mediator.start_consuming()
        except KeyboardInterrupt:
            self.echo.log_warning('interrupted')
            exit(0)



import json
import os.path
from argparse import ArgumentParser, ArgumentTypeError


CWD = os.getcwdb()


def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    raise ArgumentTypeError('Boolean value expected.')


def file(v):
    if os.path.isfile(v):
        return v
    rel = os.path.join(CWD, os.path.normpath(v))
    if os.path.isfile(rel):
        return rel
    raise ArgumentTypeError(f'file {v} is not exists')


def json_data(v):
    try:
        return json.loads(v)
    except Exception as e:
        raise ArgumentTypeError(f'INVALID JSON: {e}')


CMD_INIT = 'init'
CMD_CHECK = 'check'
CMD_CONSUME = 'consume'
CMD_PRODUCE = 'produce'
CMD_CRON = 'cron'
CMD_SCAFFOLD = 'scaffold'


parser = ArgumentParser(description='UNIPIPELINE: simple way to build the declarative and distributed data pipelines. this is cli tool for unipipeline')
parser.usage = 'unipipeline --help'

DEFAULT_CONFIG_FILE = './unipipeline.yml'
parser.add_argument('--config-file', '-f', default=DEFAULT_CONFIG_FILE, type=file, dest='config_file', help=f'path to unipipeline config file (default: {DEFAULT_CONFIG_FILE})', required=True)
parser.add_argument('--verbose', default=False, type=str2bool, const=True, nargs='?', dest='verbose', help='verbose output (default: false)')

subparsers = parser.add_subparsers(help='sub-commands', required=True, dest='cmd')

check_help = 'check loading of all modules'
check_parser = subparsers.add_parser(CMD_CHECK, help=check_help)
check_parser.description = check_help
check_parser.usage = f'''
    unipipeline -f {DEFAULT_CONFIG_FILE} {CMD_CHECK}
    unipipeline -f {DEFAULT_CONFIG_FILE} --verbose=yes {CMD_CHECK}
'''

scaffold_help = 'create all modules and classes if it is absent. no args'
scaffold_parser = subparsers.add_parser(CMD_SCAFFOLD, help=scaffold_help)
scaffold_parser.description = scaffold_help
scaffold_parser.usage = f'''
    unipipeline -f {DEFAULT_CONFIG_FILE} {CMD_SCAFFOLD}
    unipipeline -f {DEFAULT_CONFIG_FILE} --verbose=yes {CMD_SCAFFOLD}
'''

init_help = 'initialize broker topics for workers'
init_parser = subparsers.add_parser(CMD_INIT, help=init_help)
init_parser.add_argument('--workers', '-w', type=str, nargs='+', default=[], required=False, dest='init_workers', help='workers list for initialization (default: [])')
init_parser.description = init_help
init_parser.usage = f'''
    unipipeline -f {DEFAULT_CONFIG_FILE} {CMD_INIT}
    unipipeline -f {DEFAULT_CONFIG_FILE} --verbose=yes {CMD_INIT}
    unipipeline -f {DEFAULT_CONFIG_FILE} --verbose=yes {CMD_INIT} --workers some_worker_name_01 some_worker_name_02
'''

consume_help = 'start consuming workers. connect to brokers and waiting for messages'
consume_parser = subparsers.add_parser(CMD_CONSUME, help=consume_help)
consume_parser.add_argument('--workers', '-w', type=str, nargs='+', required=True, dest='consume_workers', help='worker list for consuming')
consume_parser.description = consume_help
consume_parser.usage = f'''
    unipipeline -f {DEFAULT_CONFIG_FILE} {CMD_CONSUME}
    unipipeline -f {DEFAULT_CONFIG_FILE} --verbose=yes {CMD_CONSUME}
    unipipeline -f {DEFAULT_CONFIG_FILE} {CMD_CONSUME} --workers some_worker_name_01 some_worker_name_02
    unipipeline -f {DEFAULT_CONFIG_FILE} --verbose=yes {CMD_CONSUME} --workers some_worker_name_01 some_worker_name_02
'''

cron_help = 'start cron jobs, That defined in config file'
cron_parser = subparsers.add_parser(CMD_CRON, help=cron_help)
cron_parser.description = cron_help
cron_parser.usage = f'''
    unipipeline -f {DEFAULT_CONFIG_FILE} {CMD_CRON}
    unipipeline -f {DEFAULT_CONFIG_FILE} --verbose=yes {CMD_CRON}
'''

produce_help = 'publish message to broker. send it to worker'
produce_parser = subparsers.add_parser(CMD_PRODUCE, help=produce_help)
produce_parser.add_argument('--alone', '-a', type=str2bool, nargs='?', const=True, default=False, dest='produce_alone', help='message will be sent only if topic is empty')
produce_parser.add_argument('--worker', '-w', type=str, required=True, dest='produce_worker', help='worker recipient')
produce_parser.add_argument('--data', '-d', type=json_data, required=True, dest='produce_data', help='data for sending')
produce_parser.description = produce_help
produce_parser.usage = f'''
    unipipeline -f {DEFAULT_CONFIG_FILE} {CMD_PRODUCE} --worker some_worker_name_01 --data {{"some": "json", "value": "for worker"}}
    unipipeline -f {DEFAULT_CONFIG_FILE} --verbose=yes {CMD_PRODUCE} --worker some_worker_name_01 --data {{"some": "json", "value": "for worker"}}
    unipipeline -f {DEFAULT_CONFIG_FILE} {CMD_PRODUCE} --alone --worker some_worker_name_01 --data {{"some": "json", "value": "for worker"}}
    unipipeline -f {DEFAULT_CONFIG_FILE} --verbose=yes {CMD_PRODUCE} --alone --worker some_worker_name_01 --data {{"some": "json", "value": "for worker"}}
'''


def parse_args():
    return parser.parse_args()



import bz2
import json
import lzma
import zlib
from typing import Dict, Any, Callable, Set, Type, TypeVar, Generic

CONTENT_TYPE__APPLICATION_JSON = 'application/json'
# CONTENT_TYPE__APPLICATION_DATA = 'application/data'

COMPRESSION__GZIP = 'application/x-gzip'
COMPRESSION__BZ2 = 'application/x-bz2'
COMPRESSION__LZMA = 'application/x-lzma'


TContentEncoding = str
TContentType = str


T = TypeVar('T')


def ensure_type(name: str, value: T, constructor: Type[T]) -> T:
    if not isinstance(value, constructor):
        raise TypeError(f'{name} has invalid type. must be {constructor.__name__}, {type(value).__name__} was given')
    return value


def prepare_content_type(content_type: TContentType) -> TContentType:
    ensure_type('content_type', content_type, TContentType)
    return content_type.lower()


TSerializedData = TypeVar("TSerializedData")
TParsedData = TypeVar("TParsedData")


class SerializersRegistry(Generic[TParsedData, TSerializedData]):
    def __init__(self) -> None:
        self._encoders: Dict[TContentType, Callable[[TParsedData], TSerializedData]] = dict()
        self._decoders: Dict[TContentType, Callable[[TSerializedData], TParsedData]] = dict()
        self._content_types: Set[TContentType] = set()

    def register(self, content_type: TContentType, encoder: Callable[[TParsedData], TSerializedData], decoder: Callable[[TSerializedData], TParsedData]) -> None:
        content_type = prepare_content_type(content_type)
        self._encoders[content_type] = encoder
        self._decoders[content_type] = decoder
        self._content_types.add(content_type)

    def supports(self, content_type: TContentType) -> bool:
        content_type = prepare_content_type(content_type)
        return content_type in self._content_types

    def assert_supports(self, content_type: TContentType) -> None:
        content_type = prepare_content_type(content_type)
        if content_type not in self._content_types:
            raise TypeError(f'content_type "{content_type}" is not supported')

    def dumps(self, data: TParsedData, content_type: TContentType) -> TSerializedData:
        content_type = prepare_content_type(content_type)
        encoder = self._encoders[content_type]
        return encoder(data)

    def loads(self, data: TSerializedData, content_type: TContentType) -> TParsedData:
        content_type = prepare_content_type(content_type)
        decoder = self._decoders[content_type]
        return decoder(data)


serializer_registry = SerializersRegistry[Any, str]()
serializer_registry.register(CONTENT_TYPE__APPLICATION_JSON, json.dumps, json.loads)

compressor_registry = SerializersRegistry[bytes, bytes]()
compressor_registry.register(COMPRESSION__GZIP, zlib.compress, zlib.decompress)
compressor_registry.register(COMPRESSION__BZ2, bz2.compress, bz2.decompress)
compressor_registry.register(COMPRESSION__LZMA, lzma.compress, lzma.decompress)



import logging
import os.path

from unipipeline import Uni


logging.basicConfig(
    level=os.environ.get('LOGLEVEL', logging.DEBUG),
    format='%(asctime)s | %(name)s | %(levelname)s | %(message)s'
)

CWD = os.path.dirname(os.path.abspath(__file__))

u = Uni(f"{CWD}/dag.yml")

u.check(create=True)

u.init_producer_worker('input_worker')

u.init_consumer_worker("input_worker")

u.init_consumer_worker("my_super_cron_worker")

u.initialize(create=True)

u.send_to("input_worker", dict())

u.start_consuming()

u.start_cron()



from uuid import UUID

from unipipeline.modules.uni_definition import UniDefinition
from unipipeline.modules.uni_worker_definition import UniWorkerDefinition


class UniCronTaskDefinition(UniDefinition):
    id: UUID
    name: str
    worker: UniWorkerDefinition
    when: str
    alone: bool
